# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

"""
Common functions and definitions in the example build environment.
"""

# Standard libraries
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from tsfpga.module_list import ModuleList

# Import before others since it modifies PYTHONPATH. pylint: disable=unused-import
import tsfpga.examples.example_pythonpath  # noqa: F401

# Third party libraries
from hdl_registers.register import Register

# First party libraries
import tsfpga
from tsfpga.module import get_modules

TSFPGA_EXAMPLES_TEMP_DIR = tsfpga.TSFPGA_GENERATED


def get_default_registers() -> list[Register]:
    """
    Default registers for tsfpga examples.
    """
    registers = [
        Register("config", 0, "r_w", "Configuration register."),
        Register(
            "command",
            1,
            "wpulse",
            "When this register is written, all '1's in the written word will be asserted for one "
            "clock cycle in the FPGA logic.",
        ),
        Register("status", 2, "r", "Status register."),
        Register(
            "irq_status",
            3,
            "r_wpulse",
            "Reading a '1' in this register means the corresponding interrupt has triggered.\n"
            "Writing to this register will clear the interrupts where there is a '1' in the "
            "written word.",
        ),
        Register(
            "irq_mask",
            4,
            "r_w",
            "A '1' in this register means that the corresponding interrupt is enabled.",
        ),
    ]
    return registers


def get_tsfpga_example_modules(
    names_include: Optional[set[str]] = None, names_avoid: Optional[set[str]] = None
) -> "ModuleList":
    """
    Wrapper of the regular :func:`.get_modules`. call with correct settings for tsfpga
    example modules.
    This will include the example tsfpga modules, but not the "real" modules.

    Arguments will be passed on to :func:`.get_modules`.
    """
    return get_modules(
        modules_folders=[tsfpga.TSFPGA_EXAMPLE_MODULES],
        names_include=names_include,
        names_avoid=names_avoid,
        library_name_has_lib_suffix=False,
        default_registers=get_default_registers(),
    )


def get_hdl_modules(
    names_include: Optional[set[str]] = None, names_avoid: Optional[set[str]] = None
) -> "ModuleList":
    """
    Wrapper of :func:`.get_modules` which returns the ``hdl-modules`` module objects
    (https://hdl-modules.com), if available.

    If ``hdl-modules`` can not be found in the default repo checkout location,
    the function will assert False.

    Arguments will be passed on to :func:`.get_modules`.

    Return:
        :class:`.ModuleList`: The module objects.
    """
    # Presumed location of the hdl-modules repo
    hdl_modules_repo_root = tsfpga.REPO_ROOT.parent.parent.resolve() / "hdl-modules" / "hdl-modules"
    if (hdl_modules_repo_root / "modules").exists():
        return get_modules(
            modules_folders=[hdl_modules_repo_root / "modules"],
            names_include=names_include,
            names_avoid=names_avoid,
            library_name_has_lib_suffix=False,
        )

    raise FileNotFoundError(
        f"The hdl-modules modules could not be found. Searched in {hdl_modules_repo_root}"
    )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from unittest.mock import MagicMock, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.ip_core_file import IpCoreFile
from tsfpga.module import BaseModule, get_modules
from tsfpga.system_utils import create_file, delete
from tsfpga.vivado.ip_cores import VivadoIpCores


def test_get_ip_core_files_is_called_with_the_correct_arguments(tmp_path):
    modules = [MagicMock(spec=BaseModule)]

    with patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True):
        VivadoIpCores(modules, tmp_path, part_name="test_part")

    modules[0].get_ip_core_files.assert_called_once_with(generics={}, part="test_part")


def test_system_call_to_vivado_failing_should_raise_exception(tmp_path):
    project = VivadoIpCores(modules=[], output_path=tmp_path, part_name="test_part")

    with patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True) as create:
        create.return_value = False

        with pytest.raises(AssertionError) as exception_info:
            project.create_vivado_project_if_needed()

        assert str(exception_info.value) == "Failed to create Vivado IP core project"


@pytest.fixture
def ip_cores_test(tmp_path):
    class IpCoresTest:
        def __init__(self):
            self.project_folder = tmp_path / "ip_project"
            self.modules_folder = tmp_path / "modules"

            self.apa_tcl = create_file(self.modules_folder / "apa" / "ip_cores" / "apa.tcl", "apa")
            self.hest_tcl = create_file(
                self.modules_folder / "hest" / "ip_cores" / "hest.tcl", "hest"
            )

            modules = get_modules([self.modules_folder])
            self.vivado_ip_cores = VivadoIpCores(modules, self.project_folder, part_name="-")

            # Create initial hash and (empty) compile order file
            self.vivado_ip_cores._save_hash()  # pylint: disable=protected-access
            self.create_compile_order_file()

        def create_compile_order_file(self):
            create_file(self.vivado_ip_cores.compile_order_file)

    return IpCoresTest()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_not_recreate(create, ip_cores_test):
    assert not ip_cores_test.vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_not_called()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_recreate_if_compile_order_file_is_missing(create, ip_cores_test):
    delete(ip_cores_test.vivado_ip_cores.compile_order_file)
    assert ip_cores_test.vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_called_once()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_recreate_if_hash_file_is_missing(create, ip_cores_test):
    delete(ip_cores_test.vivado_ip_cores._hash_file)  # pylint: disable=protected-access
    assert ip_cores_test.vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_called_once()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_not_recreate_if_nothing_is_changed(create, ip_cores_test):
    # This test shows that the pattern used in the upcoming tests:
    #   change something -> get modules -> create new VivadoIpCores object
    # should not result in a recreate unless we actually change something.
    modules = get_modules([ip_cores_test.modules_folder])
    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert not vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_not_called()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_recreate_if_ip_core_file_is_added(create, ip_cores_test):
    create_file(ip_cores_test.modules_folder / "zebra" / "ip_cores" / "zebra.tcl", "zebra")
    modules = get_modules([ip_cores_test.modules_folder])
    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_called_once()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_recreate_if_ip_core_file_is_removed(create, ip_cores_test):
    delete(ip_cores_test.hest_tcl)
    modules = get_modules([ip_cores_test.modules_folder])
    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_called_once()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_should_recreate_if_ip_core_file_is_changed(create, ip_cores_test):
    create_file(ip_cores_test.apa_tcl, "blaha blaha")
    modules = get_modules([ip_cores_test.modules_folder])
    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert vivado_ip_cores.create_vivado_project_if_needed()
    create.assert_called_once()


@patch("tsfpga.vivado.ip_cores.VivadoIpCoreProject.create", autospec=True)
def test_ip_core_variables(create, ip_cores_test):
    modules = [MagicMock(spec=BaseModule)]

    modules[0].get_ip_core_files.return_value = [IpCoreFile(path=ip_cores_test.apa_tcl)]

    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert vivado_ip_cores.create_vivado_project_if_needed()
    assert create.call_count == 1
    ip_cores_test.create_compile_order_file()

    # Should not recreate until we change something
    assert not vivado_ip_cores.create_vivado_project_if_needed()
    assert create.call_count == 1

    # Adding variables should recreate
    modules[0].get_ip_core_files.return_value = [
        IpCoreFile(path=ip_cores_test.apa_tcl, zz="123", aa="456")
    ]

    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert vivado_ip_cores.create_vivado_project_if_needed()
    assert create.call_count == 2
    ip_cores_test.create_compile_order_file()

    # Changing order of variables should not recreate
    modules[0].get_ip_core_files.return_value = [
        IpCoreFile(path=ip_cores_test.apa_tcl, aa="456", zz="123")
    ]

    vivado_ip_cores = VivadoIpCores(modules, ip_cores_test.project_folder, part_name="-")

    assert not vivado_ip_cores.create_vivado_project_if_needed()
    assert create.call_count == 2



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from unittest.mock import patch

# First party libraries
from tsfpga.vivado.common import get_git_sha_slv, get_vivado_version, run_vivado_tcl

THIS_DIR = Path(__file__).parent


def test_run_vivado_tcl():
    vivado_path = THIS_DIR / "vivado.exe"
    tcl_file = THIS_DIR / "script.tcl"
    expected_cmd = [
        str(vivado_path.resolve()),
        "-mode",
        "batch",
        "-notrace",
        "-source",
        str(tcl_file.resolve()),
    ]

    with patch("tsfpga.vivado.common.Process") as mocked_process:
        mocked_process.NonZeroExitCode = ValueError
        assert run_vivado_tcl(vivado_path, tcl_file)
        mocked_process.assert_called_once_with(args=expected_cmd, cwd=THIS_DIR)

    with patch("tsfpga.vivado.common.Process") as mocked_process:
        mocked_process.NonZeroExitCode = ValueError
        assert run_vivado_tcl(vivado_path, tcl_file, no_log_file=True)
        mocked_process.assert_called_once_with(
            args=expected_cmd + ["-nojournal", "-nolog"], cwd=THIS_DIR
        )

    with patch("tsfpga.vivado.common.Process") as mocked_process:
        mocked_process.NonZeroExitCode = ValueError
        mocked_process.return_value.consume_output.side_effect = ValueError("Non-zero exit code!")
        assert not run_vivado_tcl(vivado_path, tcl_file, no_log_file=True)


def test_get_vivado_version():
    assert (
        get_vivado_version(vivado_path=Path("/home/lukas/work/Xilinx/Vivado/2021.2/bin/vivado"))
        == "2021.2"
    )


def test_get_git_sha_slv():
    with patch("tsfpga.vivado.common.get_git_sha") as get_git_sha:
        get_git_sha.return_value = "abcdef0123456789"
        assert get_git_sha_slv(git_directory=None) == (
            "10101011110011011110111100000001",
            "00100011010001010110011110001001",
        )

    with patch("tsfpga.vivado.common.get_git_sha") as get_git_sha:
        get_git_sha.return_value = "00abcdef00123400"
        assert get_git_sha_slv(git_directory=None) == (
            "00000000101010111100110111101111",
            "00000000000100100011010000000000",
        )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.build_step_tcl_hook import BuildStepTclHook
from tsfpga.constraint import Constraint
from tsfpga.module import BaseModule
from tsfpga.system_utils import create_directory, create_file
from tsfpga.vivado.generics import StringGenericValue
from tsfpga.vivado.project import VivadoNetlistProject, VivadoProject, copy_and_combine_dicts


def test_casting_to_string():
    project = VivadoProject(name="my_project", modules=[], part="")
    assert (
        str(project)
        == """\
my_project
Type:       VivadoProject
Top level:  my_project_top
Generics:   -
"""
    )

    project = VivadoProject(
        name="my_project",
        modules=[],
        part="",
        top="apa",
        generics=dict(hest=True, zebra=3, foo=StringGenericValue("/home/test.vhd")),
    )
    assert (
        str(project)
        == """\
my_project
Type:       VivadoProject
Top level:  apa
Generics:   hest=True, zebra=3, foo=/home/test.vhd
"""
    )

    project = VivadoProject(name="my_project", modules=[], part="", apa=123, hest=456)
    assert (
        str(project)
        == """\
my_project
Type:       VivadoProject
Top level:  my_project_top
Generics:   -
Arguments:  apa=123, hest=456
"""
    )


def test_modules_list_should_be_copied():
    modules = [1]
    proj = VivadoProject(name="name", modules=modules, part="part")

    modules.append(2)
    assert len(proj.modules) == 1


def test_static_generics_dictionary_should_be_copied():
    generics = dict(apa=3)
    proj = VivadoProject(name="name", modules=[], part="part", generics=generics)

    generics["apa"] = False
    assert proj.static_generics["apa"] == 3


def test_constraints_list_should_be_copied():
    constraints = [Constraint(file="1")]
    proj = VivadoProject(name="name", modules=[], part="part", constraints=constraints)

    constraints.append(Constraint(file="2"))
    assert len(proj.constraints) == 1


def test_bad_constraint_type_should_raise_error():
    # Correct type should not give error
    VivadoProject(name="name", modules=[], part="part", constraints=[Constraint(file=None)])

    # Bad type should raise exception
    with pytest.raises(TypeError) as exception_info:
        VivadoProject(name="name", modules=[], part="part", constraints=["file.vhd"])
    assert str(exception_info.value) == 'Got bad type for "constraints" element: file.vhd'


def test_bad_tcl_sources_type_should_raise_error():
    # Correct type should not give error
    VivadoProject(name="name", modules=[], part="part", tcl_sources=[Path()])

    # Bad type should raise exception
    with pytest.raises(TypeError) as exception_info:
        VivadoProject(name="name", modules=[], part="part", tcl_sources=["file.tcl"])
    assert str(exception_info.value) == 'Got bad type for "tcl_sources" element: file.tcl'


def test_bad_build_step_hooks_type_should_raise_error():
    # Correct type should not give error
    VivadoProject(
        name="name",
        modules=[],
        part="part",
        build_step_hooks=[BuildStepTclHook(tcl_file="", hook_step="")],
    )

    # Bad type should raise exception
    with pytest.raises(TypeError) as exception_info:
        VivadoProject(name="name", modules=[], part="part", build_step_hooks=["file.tcl"])
    assert str(exception_info.value) == 'Got bad type for "build_step_hooks" element: file.tcl'


def test_create_should_raise_exception_if_project_path_already_exists(tmp_path):
    create_directory(tmp_path / "projects" / "name")
    proj = VivadoProject(name="name", modules=[], part="part")
    with pytest.raises(ValueError) as exception_info:
        proj.create(tmp_path / "projects")
    assert str(exception_info.value).startswith("Folder already exists")


def test_build_should_raise_exception_if_project_does_not_exists(tmp_path):
    create_directory(tmp_path / "projects")
    proj = VivadoProject(name="name", modules=[], part="part")
    with pytest.raises(ValueError) as exception_info:
        proj.build(tmp_path / "projects", synth_only=True)
    assert str(exception_info.value).startswith("Project file does not exist")


def test_build_with_impl_run_should_raise_exception_if_no_output_path_is_given():
    proj = VivadoProject(name="name", modules=[], part="part")
    with pytest.raises(ValueError) as exception_info:
        proj.build("None")
    assert str(exception_info.value).startswith("Must specify output_path")


def test_top_name():
    assert VivadoProject(name="apa", modules=[], part="").top == "apa_top"
    assert VivadoProject(name="apa", modules=[], part="", top="hest").top == "hest"


def test_project_file_name_is_same_as_project_name():
    project_path = Path("projects/apa")
    assert (
        VivadoProject(name="apa", modules=[], part="").project_file(project_path)
        == project_path / "apa.xpr"
    )


def test_project_create(tmp_path):
    with patch("tsfpga.vivado.project.run_vivado_tcl", autospec=True) as _:
        assert VivadoProject(name="apa", modules=[], part="").create(tmp_path / "projects" / "apa")
    assert (tmp_path / "projects" / "apa" / "create_vivado_project.tcl").exists()


def test_project_create_should_raise_exception_if_project_path_already_exists(tmp_path):
    project_path = create_directory(tmp_path / "projects" / "apa")
    with pytest.raises(ValueError) as exception_info:
        VivadoProject(name="apa", modules=[], part="").create(project_path)
    assert str(exception_info.value) == f"Folder already exists: {project_path}"


def test_copy_and_combine_dict_with_both_arguments_none():
    assert copy_and_combine_dicts(None, None) == {}


def test_copy_and_combine_dict_with_first_argument_valid():
    dict_first = dict(first=1)

    result = copy_and_combine_dicts(dict_first, None)
    assert result == dict(first=1)
    assert dict_first == dict(first=1)

    dict_first["first_dummy"] = True
    assert result == dict(first=1)


def test_copy_and_combine_dict_with_second_argument_valid():
    dict_second = dict(second=2)

    result = copy_and_combine_dicts(None, dict_second)
    assert result == dict(second=2)
    assert dict_second == dict(second=2)

    dict_second["second_dummy"] = True
    assert result == dict(second=2)


def test_copy_and_combine_dict_with_both_arguments_valid():
    dict_first = dict(first=1)
    dict_second = dict(second=2)

    result = copy_and_combine_dicts(dict_first, dict_second)
    assert result == dict(first=1, second=2)
    assert dict_first == dict(first=1)
    assert dict_second == dict(second=2)

    dict_first["first_dummy"] = True
    dict_second["second_dummy"] = True
    assert result == dict(first=1, second=2)


def test_copy_and_combine_dict_with_both_arguments_valid_and_same_key():
    dict_first = dict(first=1, common=3)
    dict_second = dict(second=2, common=4)

    result = copy_and_combine_dicts(dict_first, dict_second)
    assert result == dict(first=1, second=2, common=4)
    assert dict_first == dict(first=1, common=3)
    assert dict_second == dict(second=2, common=4)

    dict_first["first_dummy"] = True
    dict_second["second_dummy"] = True
    assert result == dict(first=1, second=2, common=4)


@pytest.fixture
def vivado_project_test(tmp_path):
    class VivadoProjectTest:  # pylint: disable=too-many-instance-attributes
        def __init__(self):
            self.project_path = tmp_path / "projects" / "apa" / "project"
            self.output_path = tmp_path / "projects" / "apa"
            self.ip_cache_path = MagicMock()
            self.build_time_generics = dict(enable=True)
            self.num_threads = 4
            self.run_index = 3
            self.synth_only = False
            self.from_impl = False

            self.mocked_run_vivado_tcl = None

        def create(self, project, **other_arguments):
            with patch(
                "tsfpga.vivado.project.run_vivado_tcl", autospec=True
            ) as self.mocked_run_vivado_tcl:
                return project.create(
                    project_path=self.project_path,
                    ip_cache_path=self.ip_cache_path,
                    **other_arguments,
                )

        def build(self, project):
            with patch(
                "tsfpga.vivado.project.run_vivado_tcl", autospec=True
            ) as self.mocked_run_vivado_tcl, patch(
                "tsfpga.vivado.project.VivadoProject._get_size", autospec=True
            ) as _, patch(
                "tsfpga.vivado.project.shutil.copy2", autospec=True
            ) as _:
                create_file(self.project_path / "apa.xpr")
                return project.build(
                    project_path=self.project_path,
                    output_path=self.output_path,
                    run_index=self.run_index,
                    generics=self.build_time_generics,
                    synth_only=self.synth_only,
                    num_threads=self.num_threads,
                    other_parameter="hest",
                )

    return VivadoProjectTest()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


def test_default_pre_create_hook_should_pass(vivado_project_test):
    class CustomVivadoProject(VivadoProject):
        pass

    project = CustomVivadoProject(name="apa", modules=[], part="")
    vivado_project_test.create(project)
    vivado_project_test.mocked_run_vivado_tcl.assert_called_once()


def test_project_pre_create_hook_returning_false_should_fail_and_not_call_vivado_run(
    vivado_project_test,
):
    class CustomVivadoProject(VivadoProject):
        def pre_create(self, **kwargs):  # pylint: disable=unused-argument
            return False

    assert not vivado_project_test.create(CustomVivadoProject(name="apa", modules=[], part=""))
    vivado_project_test.mocked_run_vivado_tcl.assert_not_called()


def test_create_should_call_pre_create_with_correct_parameters(vivado_project_test):
    project = VivadoProject(name="apa", modules=[], part="", generics=dict(apa=123), hest=456)
    with patch("tsfpga.vivado.project.VivadoProject.pre_create") as mocked_pre_create:
        vivado_project_test.create(project, zebra=789)
    mocked_pre_create.assert_called_once_with(
        project_path=vivado_project_test.project_path,
        ip_cache_path=vivado_project_test.ip_cache_path,
        part="",
        generics=dict(apa=123),
        hest=456,
        zebra=789,
    )
    vivado_project_test.mocked_run_vivado_tcl.assert_called_once()


def test_build_module_pre_build_hook_and_create_regs_are_called(vivado_project_test):
    project = VivadoProject(
        name="apa",
        modules=[MagicMock(spec=BaseModule), MagicMock(spec=BaseModule)],
        part="",
        apa=123,
    )
    build_result = vivado_project_test.build(project)
    assert build_result.success

    for module in project.modules:
        module.pre_build.assert_called_once_with(
            project=project,
            other_parameter="hest",
            apa=123,
            project_path=vivado_project_test.project_path,
            output_path=vivado_project_test.output_path,
            run_index=vivado_project_test.run_index,
            generics=vivado_project_test.build_time_generics,
            synth_only=vivado_project_test.synth_only,
            from_impl=vivado_project_test.from_impl,
            num_threads=vivado_project_test.num_threads,
        )
        module.create_register_synthesis_files.assert_called_once()
        module.create_register_simulation_files.assert_not_called()


def test_default_pre_and_post_build_hooks_should_pass(vivado_project_test):
    class CustomVivadoProject(VivadoProject):
        pass

    build_result = vivado_project_test.build(CustomVivadoProject(name="apa", modules=[], part=""))
    assert build_result.success
    vivado_project_test.mocked_run_vivado_tcl.assert_called_once()


def test_project_pre_build_hook_returning_false_should_fail_and_not_call_vivado_run(
    vivado_project_test,
):
    class CustomVivadoProject(VivadoProject):
        def pre_build(self, **kwargs):  # pylint: disable=unused-argument
            return False

    build_result = vivado_project_test.build(CustomVivadoProject(name="apa", modules=[], part=""))
    assert not build_result.success
    vivado_project_test.mocked_run_vivado_tcl.assert_not_called()


def test_project_post_build_hook_returning_false_should_fail(vivado_project_test):
    class CustomVivadoProject(VivadoProject):
        def post_build(self, **kwargs):  # pylint: disable=unused-argument
            return False

    build_result = vivado_project_test.build(CustomVivadoProject(name="apa", modules=[], part=""))
    assert not build_result.success
    vivado_project_test.mocked_run_vivado_tcl.assert_called_once()


def test_project_build_hooks_should_be_called_with_correct_parameters(vivado_project_test):
    project = VivadoProject(
        name="apa", modules=[], part="", generics=dict(static_generic=2), apa=123
    )
    with patch("tsfpga.vivado.project.VivadoProject.pre_build") as mocked_pre_build, patch(
        "tsfpga.vivado.project.VivadoProject.post_build"
    ) as mocked_post_build:
        vivado_project_test.build(project)

    arguments = dict(
        project_path=vivado_project_test.project_path,
        output_path=vivado_project_test.output_path,
        run_index=vivado_project_test.run_index,
        generics=copy_and_combine_dicts(
            dict(static_generic=2), vivado_project_test.build_time_generics
        ),
        synth_only=vivado_project_test.synth_only,
        from_impl=vivado_project_test.from_impl,
        num_threads=vivado_project_test.num_threads,
        other_parameter="hest",
        apa=123,
    )
    mocked_pre_build.assert_called_once_with(**arguments)

    arguments.update(build_result=unittest.mock.ANY)
    mocked_post_build.assert_called_once_with(**arguments)


def test_module_pre_build_hook_returning_false_should_fail_and_not_call_vivado(vivado_project_test):
    module = MagicMock(spec=BaseModule)
    module.name = "whatever"
    project = VivadoProject(name="apa", modules=[module], part="")

    project.modules[0].pre_build.return_value = True
    build_result = vivado_project_test.build(project)
    assert build_result.success
    vivado_project_test.mocked_run_vivado_tcl.assert_called_once()

    project.modules[0].pre_build.return_value = False
    build_result = vivado_project_test.build(project)
    assert not build_result.success
    vivado_project_test.mocked_run_vivado_tcl.assert_not_called()


@patch("tsfpga.vivado.project.VivadoTcl", autospec=True)
def test_different_generic_combinations(mocked_vivado_tcl, vivado_project_test):
    mocked_vivado_tcl.return_value.build.return_value = ""

    # No generics
    vivado_project_test.build_time_generics = None
    build_result = vivado_project_test.build(VivadoProject(name="apa", modules=[], part=""))
    assert build_result.success
    # Note: In python 3.8 we can use call_args.kwargs straight away
    _, kwargs = mocked_vivado_tcl.return_value.build.call_args
    assert kwargs["generics"] == {}

    # Only build time generics
    vivado_project_test.build_time_generics = dict(runtime="value")
    build_result = vivado_project_test.build(VivadoProject(name="apa", modules=[], part=""))
    assert build_result.success
    _, kwargs = mocked_vivado_tcl.return_value.build.call_args
    assert kwargs["generics"] == dict(runtime="value")

    # Static and build time generics
    vivado_project_test.build_time_generics = dict(runtime="value")
    build_result = vivado_project_test.build(
        VivadoProject(name="apa", modules=[], part="", generics=dict(static="a value"))
    )
    assert build_result.success
    _, kwargs = mocked_vivado_tcl.return_value.build.call_args
    assert kwargs["generics"] == dict(runtime="value", static="a value")

    # Same key in both static and build time generic. Should prefer build time.
    vivado_project_test.build_time_generics = dict(static_and_runtime="build value")
    build_result = vivado_project_test.build(
        VivadoProject(
            name="apa", modules=[], part="", generics=dict(static_and_runtime="static value")
        )
    )
    assert build_result.success
    _, kwargs = mocked_vivado_tcl.return_value.build.call_args
    assert kwargs["generics"] == dict(static_and_runtime="build value")

    # Only static generics
    vivado_project_test.build_time_generics = None
    build_result = vivado_project_test.build(
        VivadoProject(name="apa", modules=[], part="", generics=dict(runtime="a value"))
    )
    assert build_result.success
    _, kwargs = mocked_vivado_tcl.return_value.build.call_args
    assert kwargs["generics"] == dict(runtime="a value")


def test_build_time_generics_are_copied(vivado_project_test):
    vivado_project_test.build_time_generics = dict(runtime="value")
    with patch("tsfpga.vivado.project.VivadoTcl", autospec=True) as mocked_vivado_tcl:
        mocked_vivado_tcl.return_value.build.return_value = ""
        build_result = vivado_project_test.build(
            VivadoProject(name="apa", modules=[], part="", generics=dict(static="a value"))
        )
    assert build_result.success
    assert vivado_project_test.build_time_generics == dict(runtime="value")


def test_modules_are_deep_copied_before_pre_create_hook(vivado_project_test):
    class CustomVivadoProject(VivadoProject):
        def pre_create(self, **kwargs):
            self.modules[0].registers = "Some other value"
            return True

    module = MagicMock(spec=BaseModule)
    module.registers = "Some value"

    project = CustomVivadoProject(name="apa", modules=[module], part="")
    assert vivado_project_test.create(project)

    assert module.registers == "Some value"


def test_modules_are_deep_copied_before_pre_build_hook(vivado_project_test):
    class CustomVivadoProject(VivadoProject):
        def pre_build(self, **kwargs):
            self.modules[0].registers = "Some other value"
            return True

    module = MagicMock(spec=BaseModule)
    module.registers = "Some value"

    project = CustomVivadoProject(name="apa", modules=[module], part="")
    assert vivado_project_test.build(project).success

    assert module.registers == "Some value"


def test_get_size_is_called_correctly(vivado_project_test):
    project = VivadoProject(name="apa", modules=[], part="")

    def _build_with_size(synth_only):
        """
        The project.build() call is very similar to _build() method in this class, but it mocks
        the _get_size() method in a different way.
        """
        with patch(
            "tsfpga.vivado.project.run_vivado_tcl", autospec=True
        ) as vivado_project_test.mocked_run_vivado_tcl, patch(
            "tsfpga.vivado.project.HierarchicalUtilizationParser.get_size", autospec=True
        ) as mocked_get_size, patch(
            "tsfpga.vivado.project.shutil.copy2", autospec=True
        ) as _:
            # Only the first return value will be used if we are in synth_only
            mocked_get_size.side_effect = ["synth_size", "impl_size"]

            build_result = project.build(
                project_path=vivado_project_test.project_path,
                output_path=vivado_project_test.output_path,
                run_index=vivado_project_test.run_index,
                synth_only=synth_only,
            )

            assert build_result.synthesis_size == "synth_size"

            if synth_only:
                mocked_get_size.assert_called_once_with("synth_file")
                assert build_result.implementation_size is None
            else:
                assert mocked_get_size.call_count == 2
                mocked_get_size.assert_any_call("synth_file")
                mocked_get_size.assert_any_call("impl_file")

                assert build_result.implementation_size == "impl_size"

    create_file(vivado_project_test.project_path / "apa.xpr")

    create_file(
        vivado_project_test.project_path / "apa.runs" / "synth_3" / "hierarchical_utilization.rpt",
        contents="synth_file",
    )
    create_file(
        vivado_project_test.project_path / "apa.runs" / "impl_3" / "hierarchical_utilization.rpt",
        contents="impl_file",
    )

    _build_with_size(synth_only=True)
    _build_with_size(synth_only=False)


def test_netlist_build_should_set_logic_level_distribution(vivado_project_test):
    def _build_with_logic_level_distribution(project):
        """
        The project.build() call is very similar to _build() method in this class, except it
        also mocks the _get_logic_level_distribution() method.
        """
        with patch(
            "tsfpga.vivado.project.run_vivado_tcl", autospec=True
        ) as vivado_project_test.mocked_run_vivado_tcl, patch(
            "tsfpga.vivado.project.VivadoProject._get_size", autospec=True
        ) as _, patch(
            "tsfpga.vivado.project.shutil.copy2", autospec=True
        ) as _, patch(
            "tsfpga.vivado.project.LogicLevelDistributionParser.get_table", autospec=True
        ) as mocked_get_table:
            mocked_get_table.return_value = "logic_table"

            build_result = project.build(
                project_path=vivado_project_test.project_path,
                output_path=vivado_project_test.output_path,
                run_index=vivado_project_test.run_index,
            )

            if project.is_netlist_build:
                mocked_get_table.assert_called_once_with("logic_file")
                assert build_result.logic_level_distribution == "logic_table"
            else:
                mocked_get_table.assert_not_called()
                assert build_result.logic_level_distribution is None
                assert build_result.maximum_logic_level is None

    create_file(vivado_project_test.project_path / "apa.xpr")
    create_file(
        vivado_project_test.project_path
        / "apa.runs"
        / "synth_3"
        / "logical_level_distribution.rpt",
        contents="logic_file",
    )

    project = VivadoNetlistProject(name="apa", modules=[], part="")
    _build_with_logic_level_distribution(project=project)

    project = VivadoProject(name="apa", modules=[], part="")
    _build_with_logic_level_distribution(project=project)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import Any

# First party libraries
from tsfpga.vivado.project import VivadoNetlistProject, VivadoProject

THIS_DIR = Path(__file__).parent


class TsfpgaExampleVivadoProject(VivadoProject):
    """
    Example Vivado project class.
    Shows how to override and extend the base behavior.
    """

    def pre_create(  # type: ignore[override]  # pylint: disable=arguments-differ
        self, generics: dict[str, Any], **kwargs: Any
    ) -> bool:
        """
        Override parent method to add custom behavior.
        Update TCL sources just before project creation.
        """
        self.tcl_sources.append(THIS_DIR / "tcl" / "example_vivado_messages.tcl")

        return super().pre_create(generics=generics, kwargs=kwargs)


class TsfpgaExampleVivadoNetlistProject(VivadoNetlistProject):
    """
    Example Vivado project class for netlist builds.
    Shows how to override and extend the base behavior.
    """

    def pre_create(  # type: ignore[override]  # pylint: disable=arguments-differ
        self, generics: dict[str, Any], **kwargs: Any
    ) -> bool:
        """
        Override parent method to add custom behavior.
        Update TCL sources just before project creation.
        """
        self.tcl_sources.append(THIS_DIR / "tcl" / "example_vivado_messages.tcl")
        self.tcl_sources.append(THIS_DIR / "tcl" / "example_vivado_netlist_messages.tcl")

        return super().pre_create(generics=generics, kwargs=kwargs)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import shutil
import sys
from pathlib import Path

# Do PYTHONPATH insert() instead of append() to prefer any local repo checkout over any pip install
THIS_DIR = Path(__file__).parent
REPO_ROOT = THIS_DIR.parent.parent.resolve()
sys.path.insert(0, str(REPO_ROOT))

# First party libraries
from tsfpga.about import REPOSITORY_URL
from tsfpga.examples.example_env import TSFPGA_EXAMPLES_TEMP_DIR, get_tsfpga_example_modules
from tsfpga.module_documentation import ModuleDocumentation
from tsfpga.system_utils import create_file
from tsfpga.tools.sphinx_doc import build_sphinx

GENERATED_SPHINX_RST = TSFPGA_EXAMPLES_TEMP_DIR / "example_sphinx_rst"
GENERATED_SPHINX_HTML = TSFPGA_EXAMPLES_TEMP_DIR / "example_sphinx_html"


def main() -> None:
    generate_documentation()

    shutil.copyfile(THIS_DIR / "conf.py", GENERATED_SPHINX_RST / "conf.py")
    build_sphinx(build_path=GENERATED_SPHINX_RST, output_path=GENERATED_SPHINX_HTML)


def generate_documentation() -> None:
    index_rst = """
Documentation of example modules
================================

.. toctree::
  :caption: Modules
  :hidden:

"""

    for module in get_tsfpga_example_modules():
        output_path = GENERATED_SPHINX_RST / "modules" / module.name

        index_rst += f"  modules/{module.name}/{module.name}\n"

        ModuleDocumentation(
            module=module,
            repository_url=f"{REPOSITORY_URL}/tree/main/{module.path.relative_to(REPO_ROOT)}",
            repository_name="GitHub",
        ).create_rst_document(output_path=output_path)

    create_file(GENERATED_SPHINX_RST / "index.rst", contents=index_rst)


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import os
from pathlib import Path

# Third party libraries
import pytest
from git import Actor, Repo

# First party libraries
from tsfpga import REPO_ROOT, TSFPGA_DOC
from tsfpga.git_utils import find_git_files, get_git_commit, git_commands_are_available
from tsfpga.system_utils import create_directory, create_file, system_is_windows

THIS_FILE = Path(__file__)
THIS_DIR = THIS_FILE.parent


def test_this_file_is_listed_by_find_git_files():
    git_files = find_git_files(directory=THIS_DIR, file_endings_include="py")
    assert THIS_FILE in git_files
    # Test with string as well as tuple
    git_files = find_git_files(directory=REPO_ROOT, file_endings_include="py")
    assert THIS_FILE in git_files
    git_files = find_git_files(directory=REPO_ROOT, file_endings_include=("py",))
    assert THIS_FILE in git_files


def test_this_file_is_not_listed_by_find_git_files_with_bad_argument():
    git_files = find_git_files(directory=TSFPGA_DOC)
    assert THIS_FILE not in git_files
    git_files = find_git_files(directory=REPO_ROOT, file_endings_include="vhd")
    assert THIS_FILE not in git_files


def test_this_file_is_not_listed_by_find_git_files_with_file_endings_avoid():
    # Test with string as well as tuple
    git_files = find_git_files(directory=REPO_ROOT, file_endings_avoid="py")
    assert THIS_FILE not in git_files
    git_files = find_git_files(directory=REPO_ROOT, file_endings_avoid=("py",))
    assert THIS_FILE not in git_files


def test_this_file_is_not_listed_by_find_git_files_with_exclude_directory():
    git_files = find_git_files(directory=REPO_ROOT, exclude_directories=[THIS_DIR])
    assert THIS_FILE not in git_files

    git_files = find_git_files(directory=REPO_ROOT, exclude_directories=[THIS_DIR.parent])
    assert THIS_FILE not in git_files

    git_files = find_git_files(directory=REPO_ROOT, exclude_directories=[THIS_DIR.parent, THIS_DIR])
    assert THIS_FILE not in git_files

    git_files = find_git_files(directory=REPO_ROOT, exclude_directories=[THIS_FILE])
    assert THIS_FILE not in git_files


def test_this_file_is_listed_by_find_git_files_with_bad_exclude_directory():
    git_files = find_git_files(directory=REPO_ROOT, exclude_directories=[THIS_DIR.parent / "apa"])
    assert THIS_FILE in git_files


def test_git_commands_are_available_should_pass():
    assert git_commands_are_available(directory=THIS_DIR)
    assert git_commands_are_available(directory=REPO_ROOT)


def test_git_commands_are_available_with_invalid_directory_should_fail():
    if system_is_windows():
        path_outside_of_repo = "c:/"
    else:
        path_outside_of_repo = "/"

    assert not git_commands_are_available(directory=path_outside_of_repo)


@pytest.fixture
def git_commit_with_local_changes_test(tmp_path):
    class TestGitCommitWithLocalChanges:
        def __init__(self):
            self.repo_path = tmp_path / "my_repo"
            self.repo = Repo.init(self.repo_path)
            self.actor = Actor("A name", "an@email.com")

            initial_file = create_file(self.repo_path / "initial_commit_file.txt")
            self.repo.index.add(str(initial_file))
            self.repo.index.commit("Initial commit", author=self.actor, committer=self.actor)

            trash_file = create_file(self.repo_path / "local_file_for_git_test.apa")
            self.repo.index.add(str(trash_file))

    return TestGitCommitWithLocalChanges()


# False positive for pytest fixtures
# pylint: disable=redefined-outer-name


def test_get_git_commit_with_local_changes(git_commit_with_local_changes_test):
    assert get_git_commit(directory=git_commit_with_local_changes_test.repo_path).endswith(
        " (local changes present)"
    )


def test_get_git_commit_with_env_variable_and_local_changes(git_commit_with_local_changes_test):
    if "GIT_COMMIT" in os.environ:
        old_env = os.environ["GIT_COMMIT"]
    else:
        old_env = None

    os.environ["GIT_COMMIT"] = "54849b5a8152b07e0809b8f90fc24d54262cb4d6"
    assert (
        get_git_commit(directory=git_commit_with_local_changes_test.repo_path)
        == f"{os.environ['GIT_COMMIT'][0:16]} (local changes present)"
    )

    if old_env is None:
        del os.environ["GIT_COMMIT"]
    else:
        os.environ["GIT_COMMIT"] = old_env


def test_get_git_commit_without_local_changes(git_commit_with_local_changes_test):
    git_commit_with_local_changes_test.repo.index.commit(
        "Trash commit",
        author=git_commit_with_local_changes_test.actor,
        committer=git_commit_with_local_changes_test.actor,
    )
    assert not get_git_commit(directory=git_commit_with_local_changes_test.repo_path).endswith(
        " (local changes present)"
    )


def test_get_git_commit_from_child_directory(git_commit_with_local_changes_test):
    child_directory = create_directory(
        git_commit_with_local_changes_test.repo_path / "child_directory"
    )
    assert get_git_commit(directory=child_directory).endswith(" (local changes present)")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import Any, Optional

# First party libraries
from tsfpga.system_utils import create_file

# Local folder libraries
from .common import run_vivado_tcl, to_tcl_path
from .simlib_common import VivadoSimlibCommon


class VivadoSimlibCommercial(VivadoSimlibCommon):
    """
    Handle Vivado simlib with a commercial simulator.
    """

    library_names = ["unisim", "secureip", "unimacro", "unifast", "xpm"]

    _tcl = (
        "set_param general.maxthreads 8\n"
        "compile_simlib "
        "-simulator {simulator_name} "
        "-simulator_exec_path {{{simulator_folder}}} "
        "-directory {{{output_path}}} "
        "-force "
        "-family all "
        "-language all "
        "-library all "
        "-no_ip_compile "
        "-no_systemc_compile "
    )

    def __init__(
        self,
        vivado_path: Optional[Path],
        output_path: Path,
        vunit_proj: Any,
        simulator_interface: Any,
    ):
        """
        Arguments:
            output_path: The compiled simlib will be placed here.
            vunit_proj: The VUnit project that is used to run simulation.
            simulator_interface: A VUnit SimulatorInterface object.
            vivado_path: Path to Vivado executable.
        """
        self._simulator_folder = Path(simulator_interface.find_prefix())
        self._simulator_name = self._get_simulator_name(simulator_interface=simulator_interface)

        super().__init__(vivado_path=vivado_path, output_path=output_path)

        self._vunit_proj = vunit_proj

    def _get_simulator_name(self, simulator_interface: Any) -> str:
        """
        Used to get the "-simulator" argument to the Vivado "compile_simlib" function.
        In some cases Vivado needs a different simulator name than what is used in VUnit.

        Arguments:
            simulator_interface: A VUnit SimulatorInterface object.

        Return:
            str: The simulator name preferred by Vivado.
        """
        # Aldec Riviera-PRO is called "rivierapro" in VUnit but Vivado needs the name "riviera"
        if simulator_interface.name == "rivierapro":
            return "riviera"

        # Siemens Questa is called "modelsim" in VUnit but Vivado needs the name "questasim".
        # See discussion in
        #   https://github.com/VUnit/vunit/issues/834
        #   https://gitlab.com/tsfpga/tsfpga/-/issues/67
        # Use the simulator installation path to decode whether we are running Questa or
        # regular ModelSim.
        if "questa" in str(self._simulator_folder).lower():
            return "questasim"

        # In other cases Vivado uses the same name as VUnit.
        # We do not do typing of the 'simulator_interface', but we know that '.name' is a string.
        return simulator_interface.name  # type: ignore[no-any-return]

    def _compile(self) -> None:
        tcl_file = self.output_path / "compile_simlib.tcl"
        tcl = self._tcl.format(
            simulator_name=self._simulator_name,
            simulator_folder=to_tcl_path(self._simulator_folder),
            output_path=to_tcl_path(self.output_path),
        )
        create_file(tcl_file, tcl)
        compile_ok = run_vivado_tcl(self._vivado_path, tcl_file)

        if not compile_ok:
            raise RuntimeError("Vivado simlib compile call failed!")

    def _get_simulator_tag(self) -> str:
        """
        Return e.g. modelsim_modeltech_pe_10_6c or riviera_riviera_pro_2018_10_x64.
        """
        simulator_version = self._simulator_folder.parent.name
        return self._format_version(f"{self._simulator_name}_{simulator_version}")

    def _add_to_vunit_project(self) -> None:
        """
        Add the compiled simlib to your VUnit project.
        """
        for library_name in self.library_names:
            library_path = self.output_path / library_name
            assert library_path.exists(), library_path

            self._vunit_proj.add_external_library(library_name, library_path)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from unittest.mock import MagicMock

# Third party libraries
import pytest
from hdl_registers.register_list import RegisterList

# First party libraries
from tsfpga.module import BaseModule
from tsfpga.module_documentation import ModuleDocumentation
from tsfpga.system_utils import create_file

# pylint: disable=redefined-outer-name


@pytest.fixture
def module_documentation(tmp_path):
    module = BaseModule(path=tmp_path / "apa", library_name="apa")
    module_documentation_object = ModuleDocumentation(module)

    data = """\
-- -------------------------------------------------------------------------------------------------
-- Copyright (c) Lukas Vik. All rights reserved.
-- -------------------------------------------------------------------------------------------------
-- Dummy from hest.vhd.
-- -------------------------------------------------------------------------------------------------

"""
    create_file(module.path / "src" / "hest.vhd", contents=data)

    return module_documentation_object


def test_documentation_header_with_no_overview_and_no_registers(module_documentation):
    rst = module_documentation.get_rst_document()

    assert "\nhest.vhd\n--------" in rst
    assert "Dummy from hest.vhd." in rst

    assert "Register interface" not in rst


def test_documentation_header_with_overview_and_registers(module_documentation):
    # pylint: disable=protected-access
    module = module_documentation._module
    module._registers = MagicMock(spec=RegisterList)

    data = "Dummy from apa.rst."
    create_file(module.path / "doc" / "apa.rst", contents=data)

    rst = module_documentation.get_rst_document()

    assert "\nhest.vhd\n--------" in rst
    assert "Dummy from hest.vhd." in rst

    assert "Dummy from apa.rst" in rst

    assert "Register interface" in rst
    assert ":download:`separate HTML page <apa_regs.html>`" in rst


def test_submodule_documentation_with_file_exclude(module_documentation):
    data = """\
-- -------------------------------------------------------------------------------------------------
-- Copyright (c) Lukas Vik. All rights reserved.
-- -------------------------------------------------------------------------------------------------
-- Dummy from excluded.vhd.
-- -------------------------------------------------------------------------------------------------

"""
    # pylint: disable=protected-access
    excluded_vhd = create_file(
        module_documentation._module.path / "rtl" / "excluded.vhd", contents=data
    )

    rst = module_documentation.get_submodule_rst(heading_character="-", heading_character_2="_")
    assert "Dummy from excluded.vhd." in rst

    rst = module_documentation.get_submodule_rst(
        heading_character="-", heading_character_2="_", exclude_files={excluded_vhd}
    )
    assert "Dummy from excluded.vhd." not in rst


def test_submodule_documentation_with_folder_exclude(module_documentation):
    data = """\
-- -------------------------------------------------------------------------------------------------
-- Copyright (c) Lukas Vik. All rights reserved.
-- -------------------------------------------------------------------------------------------------
-- Dummy from excluded.vhd.
-- -------------------------------------------------------------------------------------------------

"""
    # pylint: disable=protected-access
    create_file(module_documentation._module.path / "rtl" / "excluded.vhd", contents=data)

    rst = module_documentation.get_submodule_rst(heading_character="-", heading_character_2="_")
    assert "Dummy from excluded.vhd." in rst

    rst = module_documentation.get_submodule_rst(
        heading_character="-",
        heading_character_2="_",
        exclude_module_folders=["rtl"],
    )
    assert "Dummy from excluded.vhd." not in rst


def test_include_sim_but_not_test_folder(module_documentation):
    data = """\
-- -------------------------------------------------------------------------------------------------
-- Copyright (c) Lukas Vik. All rights reserved.
-- -------------------------------------------------------------------------------------------------
-- Dummy from bfm.vhd.
-- -------------------------------------------------------------------------------------------------

"""
    # pylint: disable=protected-access
    create_file(module_documentation._module.path / "sim" / "bfm.vhd", contents=data)

    data = """\
-- -------------------------------------------------------------------------------------------------
-- Copyright (c) Lukas Vik. All rights reserved.
-- -------------------------------------------------------------------------------------------------
-- Dummy from tb.vhd.
-- -------------------------------------------------------------------------------------------------

"""
    # pylint: disable=protected-access
    create_file(module_documentation._module.path / "test" / "tb.vhd", contents=data)

    rst = module_documentation.get_submodule_rst(heading_character="-", heading_character_2="_")
    assert "Dummy from bfm.vhd." in rst
    assert "Dummy from tb.vhd." not in rst

    rst = module_documentation.get_submodule_rst(
        heading_character="-", heading_character_2="_", exclude_module_folders=["sim"]
    )
    assert "Dummy from bfm.vhd." not in rst
    assert "Dummy from tb.vhd." not in rst



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import sys
from pathlib import Path

# Third party libraries
from setuptools import find_packages, setup

# Do PYTHONPATH insert() instead of append() to prefer any local repo checkout over any pip install
REPO_ROOT = Path(__file__).parent.resolve()
sys.path.insert(0, str(REPO_ROOT))

# First party libraries
import tsfpga
from tsfpga.about import REPOSITORY_URL, WEBSITE_URL, get_readme_rst, get_short_slogan
from tsfpga.system_utils import path_relative_to

REQUIREMENTS_TXT = tsfpga.TSFPGA_PATH / "requirements.txt"
REQUIREMENTS_DEVELOP_TXT = tsfpga.TSFPGA_PATH / "requirements_develop.txt"


def main():
    """
    Be extremely careful when making changes to this setup script.
    It is hard to see what is actually included and what is missing.
    Also the package data, and where it gets placed in the release tree, is very messy.

    When making changes it is recommended to try the release locally before committing to main.
    To test in a docker image do, e.g:

    python3 -m build
    docker run --rm --interactive --tty --volume $(pwd)/dist:/dist:ro --workdir /dist \
        python:3.11-slim-buster /bin/bash
    python -m pip install tsfpga-*.tar.gz

    The install should pass and you should be able to run python and "import tsfpga".
    You should see all the files in "/usr/local/lib/python3.11/site-packages/tsfpga".
    Check that e.g. pylintrc, vivado_settings.tcl, module_fifo.py, etc. are there.
    Test to run "python -m pip uninstall tsfpga" and see that it passes.
    Check the output to see that there are not package files installed in weird locations
    (such as /usr/local/lib/).

    Can also try to run unit tests and simulation/build in the hdl-registers and hdl-modules
    repositories with the newly created release dist installed.
    If that passes, then everything that is needed is probably included.
    """
    # Make sure the "tests" folder does not get included in the release.
    packages = find_packages(include=["tsfpga", "tsfpga.*"])
    for package_name in packages:
        assert package_name.startswith("tsfpga"), package_name

    setup(
        name="tsfpga",
        #
        version=tsfpga.__version__,
        #
        description=get_short_slogan(),
        #
        long_description=get_readme_rst(include_extra_for_pypi=True),
        long_description_content_type="text/x-rst",
        #
        author="Lukas Vik",
        author_email="10241915+LukasVik@users.noreply.github.com",
        #
        packages=packages,
        package_data={"tsfpga": get_package_data()},
        #
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Information Technology",
            "License :: OSI Approved :: BSD License",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: OS Independent",
            "Operating System :: POSIX :: Linux",
            "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
            "Topic :: Scientific/Engineering",
            "Topic :: Software Development :: Testing",
            "Topic :: Software Development",
        ],
        #
        license="BSD 3-Clause License",
        # Same as on GitHub
        keywords=[
            "asic",
            "fpga",
            "ci",
            "simulation",
            "test",
            "vhdl",
            "build-automation",
            "eda",
            "test-automation",
            "rtl",
            "verilog",
            "xilinx",
            "synthesis",
            "vivado",
            "systemverilog",
            "implementation",
            "vunit",
        ],
        #
        install_requires=read_requirements_file(REQUIREMENTS_TXT),
        extras_require=dict(develop=read_requirements_file(REQUIREMENTS_DEVELOP_TXT)),
        #
        python_requires=">=3.9",
        #
        project_urls={
            "Homepage": WEBSITE_URL,
            "Documentation": WEBSITE_URL,
            "Changelog": f"{WEBSITE_URL}/release_notes.html",
            "Repository": REPOSITORY_URL,
            "Issues": f"{REPOSITORY_URL}/issues",
        },
    )


def read_requirements_file(path):
    requirements = []
    with open(path, encoding=tsfpga.DEFAULT_FILE_ENCODING) as file_handle:
        # Requirements file contains one package name per line
        for line_data in file_handle.readlines():
            if line_data:
                requirements.append(line_data.strip())

    return requirements


def get_package_data():
    """
    Get all files that shall be include with the release, apart from the package python files
    that are already there.

    Note that this will include all matching files in the file system. Highly recommended to run
    from a clean repo.
    """
    files = get_package_files(tsfpga.TSFPGA_PATH)

    # Specify path relative to the tsfpga python package folder
    package_data = [str(path_relative_to(file_path, tsfpga.TSFPGA_PATH)) for file_path in files]

    return package_data


def get_package_files(folder):
    """
    Find non-python files and non-package python files (namely, module_*.py) to include.
    """
    assert folder.exists(), folder

    all_files_and_folders = folder.glob("**/*")
    non_python_files = []
    for path in all_files_and_folders:
        if path.is_file() and path.suffix != ".py" and path.suffix != ".pyc":
            non_python_files.append(path)

    all_python_files = folder.glob("**/*.py")
    non_package_python_files = []
    for python_file in all_python_files:
        if not (python_file.parent / "__init__.py").exists():
            non_package_python_files.append(python_file)

    return non_python_files + non_package_python_files


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Third party libraries
import pytest

# First party libraries
from tsfpga.vivado.generics import (
    BitVectorGenericValue,
    StringGenericValue,
    get_vivado_tcl_generic_value,
)


def test_boolean_generics():
    assert get_vivado_tcl_generic_value(True) == "1'b1"
    assert get_vivado_tcl_generic_value(False) == "1'b0"


def test_integer_generics():
    assert get_vivado_tcl_generic_value(123) == "123"
    assert get_vivado_tcl_generic_value(0) == "0"
    assert get_vivado_tcl_generic_value(-7) == "-7"


def test_float_generics():
    assert get_vivado_tcl_generic_value(3.1415) == "3.1415"
    assert get_vivado_tcl_generic_value(0.0) == "0.0"
    assert get_vivado_tcl_generic_value(-1.72) == "-1.72"


def test_bit_vector_generics():
    assert get_vivado_tcl_generic_value(BitVectorGenericValue("1010")) == "4'b1010"

    with pytest.raises(ValueError) as exception_info:
        BitVectorGenericValue(1010)
    assert str(exception_info.value) == (
        "Expected BitVectorGenericValue value to be of type str."
        ' Got type="<class \'int\'>", value="1010".'
    )

    with pytest.raises(ValueError) as exception_info:
        BitVectorGenericValue("102A")
    assert (
        str(exception_info.value)
        == 'Expected BitVectorGenericValue value to contain only "1" or "0". Got "102A".'
    )


def test_string_generics():
    assert get_vivado_tcl_generic_value(StringGenericValue("apa")) == '"apa"'

    with pytest.raises(ValueError) as exception_info:
        StringGenericValue(3)
    assert str(exception_info.value) == (
        "Expected StringGenericValue value to be of type str."
        ' Got type="<class \'int\'>", value="3".'
    )

    with pytest.raises(ValueError) as exception_info:
        StringGenericValue("apa hest")
    assert (
        str(exception_info.value)
        == 'Expected StringGenericValue value to not contain spaces. Got "apa hest".'
    )


def test_unsupported_generic_type():
    with pytest.raises(ValueError) as exception_info:
        get_vivado_tcl_generic_value(dict(name="value"))
    assert (
        str(exception_info.value)
        == "Unsupported type for generic. Got type=\"<class 'dict'>\", value=\"{'name': 'value'}\"."
    )

    with pytest.raises(ValueError) as exception_info:
        get_vivado_tcl_generic_value("/home/test.vhd")
    assert str(exception_info.value) == (
        'Unsupported type for generic. Got type="<class \'str\'>", value="/home/test.vhd".'
        " Please use either of the explicit types StringGenericValue or BitVectorGenericValue."
    )

    with pytest.raises(ValueError) as exception_info:
        get_vivado_tcl_generic_value("01101")
    assert str(exception_info.value) == (
        'Unsupported type for generic. Got type="<class \'str\'>", value="01101".'
        " Please use either of the explicit types StringGenericValue or BitVectorGenericValue."
    )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path

# Third party libraries
import pytest

# First party libraries
from tsfpga import DEFAULT_FILE_ENCODING
from tsfpga.system_utils import create_file


def open_file_with_encoding(file: Path) -> None:
    """
    Try to open the ``file`` with ASCII encoding.
    If that fails, i.e. if it contains non-ASCII characters, print information and raise exception.
    """
    try:
        with open(file, encoding="ascii") as file_handle:
            file_handle.read()

    except UnicodeDecodeError as exception:
        print(file)
        with open(file, encoding="utf8") as file_handle:
            lines = file_handle.readlines()

        for line_idx, line in enumerate(lines):
            for character in line:
                try:
                    character.encode("ascii")
                except UnicodeEncodeError:
                    print(f"Character {character} on line {line_idx + 1} is not ASCII")

        raise exception


def check_file_ends_with_newline(file: Path) -> bool:
    """
    Return True if ``file`` ends with newline.
    """
    test_ok = True
    with open(file, encoding=DEFAULT_FILE_ENCODING) as file_handle:
        file_data = file_handle.read()
        if len(file_data) != 0:
            if file_data[-1] != "\n":
                print(f"File {file} didn't end with newline")
                test_ok = False

    return test_ok


def check_file_for_tab_character(file: Path) -> bool:
    """
    Return True of ``file`` does not contain any TAB character.
    """
    test_ok = True
    with open(file, encoding=DEFAULT_FILE_ENCODING) as file_handle:
        for idx, line in enumerate(file_handle.readlines()):
            if "\t" in line:
                test_ok = False
                print(f"TAB character (\\t) on line {idx + 1} in {file}")
    return test_ok


def check_file_for_carriage_return(file: Path) -> bool:
    """
    Return True if ``file`` does not contain any carriage return (CR).
    """
    test_ok = True
    with open(file, encoding=DEFAULT_FILE_ENCODING, newline="") as file_handle:
        if "\r" in file_handle.read():
            test_ok = False
            print(f"Windows style line breaks (\\r\\n aka CR/LF) in {file}")

    return test_ok


def check_file_for_trailing_whitespace(file: Path) -> bool:
    """
    Return True if ``file`` does not contain any trailing whitespace.
    """
    test_ok = True
    with open(file, encoding=DEFAULT_FILE_ENCODING) as file_handle:
        for idx, line in enumerate(file_handle.readlines()):
            if " \n" in line:
                test_ok = False
                print(f"Trailing whitespace on line {idx + 1} in {file}")

    return test_ok


def check_file_for_line_length(file_path: Path, max_length: int = 100) -> bool:
    """
    Return True if ``file`` does not contain any line that is longer than ``max_length`` characters.
    """
    max_length_with_newline = max_length + 1
    test_ok = True

    with open(file_path, encoding=DEFAULT_FILE_ENCODING) as file_handle:
        lines = file_handle.readlines()
        for line_number, line in enumerate(lines):
            line_length = len(line)
            if line_length > max_length_with_newline:
                print(
                    f"Line {file_path}:{line_number + 1} is too long "
                    f"({line_length - 1} > {max_length_with_newline - 1})"
                )
                test_ok = False

    return test_ok


def test_open_file_with_encoding_should_raise_exception_on_bad_file(tmp_path: Path) -> None:
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    file = tmp_path / "temp_file_for_test.txt"
    with open(file, "w", encoding="utf-8") as file_handle:
        # Swedish word for island = non-ASCII character
        data = "\N{LATIN CAPITAL LETTER O WITH DIAERESIS}"
        file_handle.write(data)

    with pytest.raises(UnicodeDecodeError):
        open_file_with_encoding(file)


def test_check_file_for_tab_character_should_fail_on_bad_file(tmp_path: Path) -> None:
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    data = "Apa\thest"
    file = create_file(tmp_path / "temp_file_for_test.txt", data)
    assert not check_file_for_tab_character(file)


def test_check_file_for_carriage_return_should_fail_on_bad_file(tmp_path: Path) -> None:
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    file = tmp_path / "temp_file_for_test.txt"
    data = b"Apa\r\nhest"
    with file.open("wb") as file_handle:
        file_handle.write(data)
    assert not check_file_for_carriage_return(file)


def test_check_file_for_trailing_whitespace(tmp_path: Path) -> None:
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    data = "Apa \nhest    \nzebra"
    file = create_file(tmp_path / "temp_file_for_test.txt", data)
    assert not check_file_for_trailing_whitespace(file)


def test_check_file_for_line_length(tmp_path: Path) -> None:
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    ok_data = """
asdf
apa
hest
"""
    ok_file_path = create_file(tmp_path / "ok_data.txt", contents=ok_data)
    assert check_file_for_line_length(ok_file_path)

    bad_data = """
asdf
apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa apa
hest
"""
    bad_file_path = create_file(tmp_path / "bad_data.txt", contents=bad_data)
    assert not check_file_for_line_length(file_path=bad_file_path, max_length=80)



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Optional

# First party libraries
from tsfpga.hdl_file import HdlFile
from tsfpga.system_utils import create_file

# Local folder libraries
from .common import to_tcl_path
from .generics import get_vivado_tcl_generic_value

if TYPE_CHECKING:
    # First party libraries
    from tsfpga.build_step_tcl_hook import BuildStepTclHook
    from tsfpga.constraint import Constraint
    from tsfpga.module_list import ModuleList


# Number of available Vivado implementation strategies
NUM_VIVADO_STRATEGIES = 33


class VivadoTcl:
    """
    Class with methods for translating a set of sources into Vivado TCL
    """

    def __init__(
        self,
        name: str,
    ) -> None:
        self.name = name

    # pylint: disable=too-many-arguments
    def create(
        self,
        project_folder: Path,
        modules: "ModuleList",
        part: str,
        top: str,
        run_index: int,
        generics: Optional[dict[str, str]] = None,
        constraints: Optional[list["Constraint"]] = None,
        tcl_sources: Optional[list[Path]] = None,
        build_step_hooks: Optional[list["BuildStepTclHook"]] = None,
        ip_cache_path: Optional[Path] = None,
        disable_io_buffers: bool = True,
        # Add no sources other than IP cores
        ip_cores_only: bool = False,
        # Will be passed on to module functions. Enables parameterization of e.g. IP cores.
        other_arguments: Optional[dict[str, Any]] = None,
    ) -> str:
        generics = {} if generics is None else generics
        other_arguments = {} if other_arguments is None else other_arguments

        tcl = f"create_project {self.name} {{{to_tcl_path(project_folder)}}} -part {part}\n"
        tcl += "set_property target_language VHDL [current_project]\n"

        if ip_cache_path is not None:
            tcl += f"config_ip_cache -use_cache_location {{{to_tcl_path(ip_cache_path)}}}\n"
            tcl += "\n"

        tcl += self._add_tcl_sources(tcl_sources)
        tcl += "\n"

        if not ip_cores_only:
            tcl += self._add_module_source_files(modules=modules, other_arguments=other_arguments)
            tcl += "\n"
            tcl += self._add_generics(generics)
            tcl += "\n"
            tcl += self._add_constraints(
                self._iterate_constraints(
                    modules=modules, constraints=constraints, other_arguments=other_arguments
                )
            )
            tcl += "\n"
            tcl += self._add_build_step_hooks(build_step_hooks, project_folder)
            tcl += "\n"

        tcl += self._add_ip_cores(modules=modules, other_arguments=other_arguments)
        tcl += "\n"
        tcl += self._add_project_settings()
        tcl += "\n"
        tcl += f"current_run [get_runs synth_{run_index}]\n"
        tcl += "\n"
        tcl += f"set_property top {top} [current_fileset]\n"
        tcl += "reorder_files -auto -disable_unused\n"
        tcl += "\n"

        if disable_io_buffers:
            tcl += (
                "set_property -name {STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS} "
                f"-value -no_iobuf -objects [get_runs synth_{run_index}]"
            )
            tcl += "\n"

        tcl += "exit\n"
        return tcl

    def _add_module_source_files(
        self, modules: "ModuleList", other_arguments: dict[str, Any]
    ) -> str:
        tcl = ""
        for module in modules:
            vhdl_files = []
            verilog_files = []
            system_verilog_files = []

            for hdl_file in module.get_synthesis_files(**other_arguments):
                if hdl_file.type == HdlFile.Type.VHDL:
                    vhdl_files.append(hdl_file.path)
                elif hdl_file.type in [HdlFile.Type.VERILOG_SOURCE, HdlFile.Type.VERILOG_HEADER]:
                    verilog_files.append(hdl_file.path)
                elif hdl_file.type in [
                    HdlFile.Type.SYSTEMVERILOG_SOURCE,
                    HdlFile.Type.SYSTEMVERILOG_HEADER,
                ]:
                    system_verilog_files.append(hdl_file.path)
                else:
                    raise NotImplementedError(f"Can not handle file: {hdl_file}")
                    # Encrypted source files (.vp?), etc, I do not know how
                    # to handle, since I have no use case for it at the moment.

            if vhdl_files:
                files_string = self._to_file_list(vhdl_files)
                tcl += f"read_vhdl -library {module.library_name} -vhdl2008 {files_string}\n"

            if verilog_files:
                files_string = self._to_file_list(verilog_files)
                tcl += f"read_verilog {files_string}\n"

            if system_verilog_files:
                files_string = self._to_file_list(system_verilog_files)
                tcl += f"read_verilog -sv {files_string}\n"

        return tcl

    @staticmethod
    def _to_file_list(file_paths: list[Path]) -> str:
        """
        Return a TCL snippet for a file list, with each file enclosed in curly braces.
        E.g. "{file1}" or "{{file1} {file2} {file3}}"
        """
        if len(file_paths) == 1:
            files_string = to_tcl_path(file_paths[0])
        else:
            files_string = " ".join([f"{{{to_tcl_path(file_path)}}}" for file_path in file_paths])

        return f"{{{files_string}}}"

    @staticmethod
    def _add_tcl_sources(tcl_sources: Optional[list[Path]]) -> str:
        if tcl_sources is None:
            return ""

        tcl = ""
        for tcl_source_file in tcl_sources:
            tcl += f"source -notrace {{{to_tcl_path(tcl_source_file)}}}\n"
        return tcl

    @staticmethod
    def _add_ip_cores(modules: "ModuleList", other_arguments: dict[str, Any]) -> str:
        tcl = ""
        for module in modules:
            for ip_core_file in module.get_ip_core_files(**other_arguments):
                create_function_name = f"create_ip_core_{ip_core_file.name}"
                tcl += f"proc {create_function_name} {{}} {{\n"

                if ip_core_file.variables:
                    for key, value in ip_core_file.variables.items():
                        tcl += f'  set {key} "{value}"\n'

                tcl += f"""\
  source -notrace {{{to_tcl_path(ip_core_file.path)}}}
}}
{create_function_name}
"""

        return tcl

    def _add_build_step_hooks(
        self, build_step_hooks: Optional[list["BuildStepTclHook"]], project_folder: Path
    ) -> str:
        if build_step_hooks is None:
            return ""

        # There can be many hooks for the same step. Reorganize them into a dict.
        hook_steps: dict[str, list["BuildStepTclHook"]] = {}
        for build_step_hook in build_step_hooks:
            if build_step_hook.hook_step in hook_steps:
                hook_steps[build_step_hook.hook_step].append(build_step_hook)
            else:
                hook_steps[build_step_hook.hook_step] = [build_step_hook]

        tcl = ""
        for step, hooks in hook_steps.items():
            # Vivado will only accept one TCL script as hook for each step. So if we want
            # to add more we have to create a new TCL file, that sources the other files,
            # and add that as the hook to Vivado.
            if len(hooks) == 1:
                tcl_file = hooks[0].tcl_file
            else:
                tcl_file = project_folder / ("hook_" + step.replace(".", "_") + ".tcl")
                source_hooks_tcl = "".join(
                    [f"source {{{to_tcl_path(hook.tcl_file)}}}\n" for hook in hooks]
                )
                create_file(tcl_file, source_hooks_tcl)

            # Add to fileset to enable archive and other project based functionality
            tcl += f"add_files -fileset utils_1 -norecurse {{{to_tcl_path(tcl_file)}}}\n"

            # Build step hook can only be applied to a run (e.g. impl_1), not on a project basis
            run_wildcard = "synth_*" if hooks[0].step_is_synth else "impl_*"
            tcl_block = f"set_property {step} {{{to_tcl_path(tcl_file)}}} ${{run}}"
            tcl += self._tcl_for_each_run(run_wildcard, tcl_block)

        return tcl

    def _add_project_settings(self) -> str:
        tcl = ""

        # Default value for when opening project in GUI.
        # Will be overwritten if using build() function.
        tcl += "set_param general.maxThreads 7\n"

        # Enable VHDL assert statements to be evaluated. A severity level of failure will
        # stop the synthesis and produce an error.
        tcl_block = "set_property STEPS.SYNTH_DESIGN.ARGS.ASSERT true ${run}"
        tcl += self._tcl_for_each_run("synth_*", tcl_block)

        # Enable binary bitstream as well
        tcl_block = "set_property STEPS.WRITE_BITSTREAM.ARGS.BIN_FILE true ${run}"
        tcl += self._tcl_for_each_run("impl_*", tcl_block)

        return tcl

    @staticmethod
    def _tcl_for_each_run(run_wildcard: str, tcl_block: str) -> str:
        """
        Apply TCL block for each defined run. Use ${run} for run variable in TCL.
        """
        tcl = ""
        tcl += f"foreach run [get_runs {run_wildcard}] {{\n"
        tcl += tcl_block + "\n"
        tcl += "}\n"
        return tcl

    @staticmethod
    def _add_generics(generics: Optional[dict[str, Any]]) -> str:
        """
        Generics are set according to this weird format:
        https://www.xilinx.com/support/answers/52217.html
        """
        if not generics:
            return ""

        generic_list = []
        for name, value in generics.items():
            value_tcl_formatted = get_vivado_tcl_generic_value(value=value)
            generic_list.append(f"{name}={value_tcl_formatted}")

        generics_string = " ".join(generic_list)
        return f"set_property generic {{{generics_string}}} [current_fileset]\n"

    @staticmethod
    def _iterate_constraints(
        modules: "ModuleList",
        constraints: Optional[list["Constraint"]],
        other_arguments: dict[str, Any],
    ) -> Iterable["Constraint"]:
        for module in modules:
            yield from module.get_scoped_constraints(**other_arguments)

        if constraints is not None:
            yield from constraints

    @staticmethod
    def _add_constraints(constraints: Iterable["Constraint"]) -> str:
        tcl = ""
        for constraint in constraints:
            constraint_file = to_tcl_path(constraint.file)

            ref_flags = "" if constraint.ref is None else (f"-ref {constraint.ref} ")
            managed_flags = "" if constraint_file.endswith("xdc") else "-unmanaged "
            tcl += f"read_xdc {ref_flags}{managed_flags}{{{constraint_file}}}\n"

            get_file = f"[get_files {{{constraint_file}}}]"
            tcl += f"set_property PROCESSING_ORDER {constraint.processing_order} {get_file}\n"

            if constraint.used_in == "impl":
                tcl += f"set_property used_in_synthesis false {get_file}\n"
            elif constraint.used_in == "synth":
                tcl += f"set_property used_in_implementation false {get_file}\n"

        return tcl

    def build(
        self,
        project_file: Path,
        output_path: Path,
        num_threads: int,
        run_index: int,
        generics: Optional[dict[str, Any]] = None,
        synth_only: bool = False,
        from_impl: bool = False,
        impl_explore: bool = False,
        analyze_synthesis_timing: bool = True,
    ) -> str:
        if impl_explore:
            # For implementation explore, threads are divided to one each per job.
            # Number of jobs in parallel are the number of threads specified for build.
            # Clamp max threads between 1 and 32, which are allowed by Vivado 2018.3+.
            num_threads_general = min(max(1, num_threads // NUM_VIVADO_STRATEGIES), 32)
        else:
            # Max value in Vivado 2018.3+. set_param will give an error if higher number.
            num_threads_general = min(num_threads, 32)

        num_threads_synth = min(num_threads, 8)

        tcl = f"open_project {to_tcl_path(project_file)}\n"
        tcl += f"set_param general.maxThreads {num_threads_general}\n"
        tcl += f"set_param synth.maxThreads {num_threads_synth}\n"
        tcl += "\n"
        tcl += self._add_generics(generics)
        tcl += "\n"

        if not from_impl:
            synth_run = f"synth_{run_index}"

            tcl += self._synthesis(synth_run, num_threads, analyze_synthesis_timing)
            tcl += "\n"

        if not synth_only:
            impl_run = f"impl_{run_index}"

            if impl_explore:
                tcl += self._run_multiple(num_jobs=num_threads)
            else:
                tcl += self._run(impl_run, num_threads, to_step="write_bitstream")
            tcl += "\n"

            tcl += self._write_hw_platform(output_path)
            tcl += "\n"

        tcl += "exit\n"

        return tcl

    def _synthesis(self, run: str, num_threads: int, analyze_synthesis_timing: bool) -> str:
        tcl = self._run(run, num_threads)
        if analyze_synthesis_timing:
            # For synthesis flow we perform the timing checks by opening the design. It would have
            # been more efficient to use a post-synthesis hook (since the design would already be
            # open), if that mechanism had worked. It seems to be very bugged. So we add the
            # checkers to the build script.
            # For implementation, we use a pre-bitstream build hook which seems to work decently.
            tcl += """
open_run ${run}
set run_directory [get_property DIRECTORY [get_runs ${run}]]

# Generate report on simultaneous switching noise (SSN) for the design.
# It seems safe to do this after synthesis; inspecting the reports in a test build after both
# synthesis and implementation shows that the results are identical.
# Will generate a "Designutils 20-923" message if noise margins are not met.
# If the user would like this to fail the build, this message severity shall be raised to ERROR.
# At the moment we do not know how stable this mechanism is, so we do not fail the build
# per default.
# The call is very fast (< 1s) so it is fine to run always, even though not everyone will use it.
set current_part [get_property PART [current_project]]
set part_supports_ssn [get_parts ${current_part} -filter {ssn_report == 1}]
if {${part_supports_ssn} != ""} {
    set output_file [file join ${run_directory} "report_ssn.html"]
    report_ssn -phase -format html -file ${output_file}
}

# This call is duplicated in report_utilization.tcl for implementation.
set output_file [file join ${run_directory} "hierarchical_utilization.rpt"]
report_utilization -hierarchical -hierarchical_depth 4 -file ${output_file}


# After synthesis we check for unhandled clock crossings and abort the build based on the result.
# Other timing checks, e.g. setup/hold/pulse width violations, are not reliable after synthesis,
# and should not abort the build. These need to be checked after implementation.
"""

            tcl += """
# This code is duplicated in check_timing.tcl for implementation.
if {[regexp {\\(unsafe\\)} [report_clock_interaction -delay_type min_max -return_string]]} {
  puts "ERROR: Unhandled clock crossing in ${run} run. See clock_interaction.rpt and \
timing_summary.rpt in ${run_directory}."

  set output_file [file join ${run_directory} "clock_interaction.rpt"]
  report_clock_interaction -delay_type min_max -file ${output_file}

  set output_file [file join ${run_directory} "timing_summary.rpt"]
  report_timing_summary -file ${output_file}

  exit 1
}
"""
        return tcl

    @staticmethod
    def _run(run: str, num_threads: int, to_step: Optional[str] = None) -> str:
        to_step = "" if to_step is None else f" -to_step {to_step}"

        tcl = f"""
set run {run}
reset_run ${{run}}
launch_runs ${{run}} -jobs {num_threads}{to_step}
"""

        tcl += """
wait_on_run ${run}

if {[get_property PROGRESS [get_runs ${run}]] != "100%"} {
  puts "ERROR: Run ${run} failed."
  exit 1
}
"""
        return tcl

    def _run_multiple(self, num_jobs: int = 4, base_name: str = "impl_explore_") -> str:
        # Currently, this creates .tcl that waits for all active runs to complete

        tcl = "set build_succeeded 0\n"
        tcl += f"reset_runs [get_runs {base_name}*]\n"
        tcl += f"launch_runs -jobs {num_jobs} [get_runs {base_name}*] -to_step write_bitstream\n"
        tcl += "\n"

        tcl += f"wait_on_runs -quiet -exit_condition ANY_ONE_MET_TIMING [get_runs {base_name}*]\n"
        tcl += "\n"

        tcl += 'reset_runs [get_runs -filter {STATUS == "Queued..."}]\n'

        # Wait on runs that are still going, since Vivado can't kill runs in progress reliably.
        # Killing runs in progress causes a zombie process which will lock up VUnit's Process class.
        tcl += f'wait_on_runs -quiet [get_runs -filter {{STATUS != "Not started"}} {base_name}*]\n'
        tcl += "\n"

        tcl_block = """
  set build_succeeded 1
  puts "Run $run met timing"
"""
        tcl += self._tcl_for_each_run(
            f'-filter {{PROGRESS == "100%"}} {base_name}*', tcl_block=tcl_block
        )

        tcl += """
if {${build_succeeded} eq 0} {
  puts "No build met timing, exiting."
  exit 1
}
"""

        return tcl

    def _write_hw_platform(self, output_path: Path) -> str:
        """
        TCL command to create a Xilinx support archive (.xsa) file, for use as a
        hardware platform.
        Used to be known as a "hdf" or "hwdef" file.

        This is mainly used for Zynq devices to generate code to set up the PS at boot.
        There is also code generated for each MicroBlaze that is present in the design.
        If there is neither a block design nor a MicroBlaze available, the .xsa will be empty apart
        from some info about which part is used, etc.

        The '-quiet' flag is used since there was a Vivado bug observed in this very
        specific scenario:
        * Vivado 2022.1
        * UltraScale+ non-Zynq device (i.e. no block design)
        * Design contains MicroBlaze
        * Design contains ILA
        In this case the 'write_hw_platform' call would fail.
        This bug might be present in other Vivado versions and in other scenarios as well.
        Since this is a very fringe scenario, and it is unlikely that anyone would ever need the
        .xsa file specifically from the ILA build, we allow the command to fail quietly.
        """
        xsa_file = to_tcl_path(output_path / f"{self.name}.xsa")

        tcl = f"""
puts "Creating hardware platform {xsa_file}..."
write_hw_platform -fixed -force -quiet -include_bit {{{xsa_file}}}
"""

        return tcl



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from unittest.mock import MagicMock, patch

# Third party libraries
import pytest

# First party libraries
from tsfpga.git_simulation_subset import GitSimulationSubset
from tsfpga.system_utils import create_file


def test_supplying_only_one_of_vunit_preprocessed_path_or_modules_should_raise_exception():
    GitSimulationSubset(repo_root=None, reference_branch=None, vunit_proj=None)
    GitSimulationSubset(
        repo_root=None,
        reference_branch=None,
        vunit_proj=None,
        vunit_preprocessed_path="dummy",
        modules="dummy",
    )

    assertion_message = "Can not supply only one of vunit_preprocessed_path and modules"

    with pytest.raises(ValueError) as exception_info:
        GitSimulationSubset(
            repo_root=None,
            reference_branch=None,
            vunit_proj=None,
            vunit_preprocessed_path="dummy",
        )
    assert str(exception_info.value) == assertion_message

    with pytest.raises(ValueError) as exception_info:
        GitSimulationSubset(
            repo_root=None,
            reference_branch=None,
            vunit_proj=None,
            modules="dummy",
        )
    assert str(exception_info.value) == assertion_message


def test_find_subset(tmp_path):  # pylint: disable=too-many-locals
    # Set up a scenario with a few files that have diffs and a few files that do not.
    # The tb without diffs will be set up so that it depends on the source file that
    # has diffs. This means that both tb's should be returned by find_subset().
    #
    # It is quite a complicated test. I am not sure that it adds a tremendous amount of value.
    # It is really just a practice in using mocks. :)
    vhd_with_diff = create_file(tmp_path / "file_with_diff.vhd")
    vhd_with_no_diff = create_file(tmp_path / "file_with_no_diff.vhd")
    tb_vhd_with_diff = create_file(tmp_path / "tb_file_with_diff.vhd")
    tb_vhd_with_no_diff = create_file(tmp_path / "tb_file_with_no_diff.vhd")

    vunit_proj = MagicMock()

    git_simulation_subset = GitSimulationSubset(
        repo_root=tmp_path, reference_branch="origin/main", vunit_proj=vunit_proj
    )

    with patch("tsfpga.git_simulation_subset.Repo", autospec=True) as mocked_repo:
        repo = mocked_repo.return_value
        head_commit = repo.head.commit

        reference_commit = repo.commit.return_value

        def diff_commit(arg):
            # Return the files that have diffs. One or the other depending on what the argument
            # (reference commit, or None=local tree) is.
            diff = MagicMock()
            if arg is None:
                diff.b_path = tb_vhd_with_diff
            elif arg is reference_commit:
                diff.b_path = vhd_with_diff
            else:
                assert False
            return [diff]

        head_commit.diff.side_effect = diff_commit

        source_file_vhd_with_diff = MagicMock()
        source_file_vhd_with_diff.name = str(vhd_with_diff)
        source_file_vhd_with_diff.library.name = "apa"

        source_file_vhd_with_no_diff = MagicMock()
        source_file_vhd_with_no_diff.name = str(vhd_with_no_diff)
        source_file_vhd_with_no_diff.library.name = "hest"

        source_file_tb_vhd_with_diff = MagicMock()
        source_file_tb_vhd_with_diff.name = str(tb_vhd_with_diff)
        source_file_tb_vhd_with_diff.library.name = "zebra"

        source_file_tb_vhd_with_no_diff = MagicMock()
        source_file_tb_vhd_with_no_diff.name = str(tb_vhd_with_no_diff)
        source_file_tb_vhd_with_no_diff.library.name = "foo"

        vunit_proj.get_source_files.return_value = [
            source_file_tb_vhd_with_diff,
            source_file_vhd_with_diff,
            source_file_vhd_with_no_diff,
            source_file_tb_vhd_with_no_diff,
        ]

        def get_implementation_subset(arg):
            if arg == [source_file_tb_vhd_with_diff]:
                # This tb, which has a diff, depends on itself and a vhd file with no diff
                return [source_file_tb_vhd_with_diff, source_file_vhd_with_no_diff]
            if arg == [source_file_tb_vhd_with_no_diff]:
                # This tb, which has no diff, depends on itself and a vhd file that does have a diff
                return [source_file_tb_vhd_with_no_diff, source_file_vhd_with_diff]
            assert False
            return None

        vunit_proj.get_implementation_subset.side_effect = get_implementation_subset

        assert git_simulation_subset.find_subset() == [
            ("tb_file_with_diff", "zebra"),
            ("tb_file_with_no_diff", "foo"),
        ]

        repo.commit.assert_called_once_with("origin/main")



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path
from typing import TYPE_CHECKING, Optional

# Third party libraries
from hdl_registers.generator.html.page import HtmlPageGenerator

# Local folder libraries
from .about import WEBSITE_URL
from .system_utils import create_file, file_is_in_directory, read_file
from .vhdl_file_documentation import VhdlFileDocumentation
from .vivado.project import VivadoNetlistProject

if TYPE_CHECKING:
    # Local folder libraries
    from .hdl_file import HdlFile
    from .module import BaseModule


class ModuleDocumentation:
    """
    Methods for generating a reStructuredText document with module documentation.
    The content is extracted from VHDL source file headers.
    """

    def __init__(
        self,
        module: "BaseModule",
        repository_url: Optional[str] = None,
        repository_name: Optional[str] = None,
    ) -> None:
        """
        Arguments:
            module: The module which shall be documented.
            repository_url: Optionally specify an URL where the source code can be viewed.
                If this argument is specified, links will be added to the documentation.
                URL should be to the module folder within the repository.
            repository_name: Optionally specify the name of the repository URL.
                For example "GitLab".
        """
        self._module = module
        self._repository_url = repository_url
        self._repository_name = repository_name

        assert (repository_url is None) == (
            repository_name is None
        ), "Both or none of the repository arguments must be set"

    def get_overview_rst(self) -> Optional[str]:
        """
        Get the contents of the module's ``doc/<name>.rst``, i.e. the module "overview" document.

        Return:
            Module overview RST. ``None`` if file does not exist.
        """
        overview_rst_file = self._module.path / "doc" / f"{self._module.name}.rst"
        if overview_rst_file.exists():
            return read_file(overview_rst_file)

        return None

    def get_register_rst(self, heading_character: str) -> Optional[str]:
        """
        Get an RST snippet with a link to the module's register documentation, if available.
        Note that this will create an RST ``:download:`` statement to the register .html page.
        When building, the ``.html`` file must be present in the same directory as the
        ``.rst`` file.
        This is done automatically by :meth:`.create_rst_document`.

        Arguments:
            heading_character: Character to use for heading underline.

        Return:
            RST snippet with link to register HTML. ``None`` if module does not have registers.
        """
        if self._module.registers is not None:
            heading = "Register interface"
            heading_underline = heading_character * len(heading)
            return f"""\
.. _{self._module.name}.register_interface:

{heading}
{heading_underline}

This module has register definitions.
Please see :download:`separate HTML page <{self._module.name}_regs.html>` for \
register documentation.
"""

        return None

    def get_submodule_rst(
        self,
        heading_character: str,
        heading_character_2: str,
        exclude_files: Optional[set[Path]] = None,
        exclude_module_folders: Optional[list[str]] = None,
    ) -> str:
        """
        Get RST code with documentation of the different sub-modules (files) of the module.
        Contains documentation that is extracted from the file headers, as well as a
        symbolator symbol of the entity.

        Arguments:
            heading_character: Character to use for heading underline.
            heading_character_2: Character to use for next level of heading underline.
            exclude_files: Files that shall be excluded from the documentation.
            exclude_module_folders: Folder names within the module root that shall be
                excluded from documentation. For example, if you chosen module structure places
                only netlist build wrappers in the "rtl/" folder within modules, and you do not
                want them included in the documentation, then pass the argument ["rtl"].

        Return:
            RST code with sub-module documentation.
        """
        exclude_module_folders = [] if exclude_module_folders is None else exclude_module_folders
        exclude_module_paths = [self._module.path / name for name in exclude_module_folders]

        all_builds = self._module.get_build_projects()

        rst = ""

        for vhdl_file_path in self._get_vhdl_files(
            exclude_files=exclude_files, exclude_folders=exclude_module_paths
        ):
            netlist_build_base_name = f"{self._module.library_name}.{vhdl_file_path.stem}"

            netlist_builds = []
            # Include all netlist builds whose project name matches this file
            for project in all_builds:
                if isinstance(project, VivadoNetlistProject) and (
                    project.name == netlist_build_base_name
                    or project.name.startswith(f"{netlist_build_base_name}.")
                ):
                    netlist_builds.append(project)

            rst += self._get_vhdl_file_rst(
                vhdl_file_path=vhdl_file_path,
                heading_character=heading_character,
                heading_character_2=heading_character_2,
                netlist_builds=netlist_builds,
            )

        return rst

    def get_rst_document(self, exclude_module_folders: Optional[list[str]] = None) -> str:
        """
        Get a complete RST document with the content of :meth:`.get_overview_rst`,
        :meth:`.get_register_rst`, and :meth:`.get_submodule_rst`, as well as a top level heading.

        Arguments:
            exclude_module_folders: Folder names within the module root that shall be
                excluded from documentation.

        Return:
            An RST document.
        """
        heading_character_1 = "="
        heading_character_2 = "-"
        heading_character_3 = "_"

        heading = f"Module {self._module.name}"
        heading_underline = heading_character_1 * len(heading)

        if self._repository_url:
            url_rst = (
                f"To browse the source code, please visit the "
                f"`repository on {self._repository_name} <{self._repository_url}>`__."
            )
        else:
            url_rst = ""

        overview_rst = self.get_overview_rst()
        overview_rst = "" if overview_rst is None else overview_rst

        registers_rst = self.get_register_rst(heading_character=heading_character_2)
        registers_rst = "" if registers_rst is None else registers_rst

        submodule_rst = self.get_submodule_rst(
            heading_character=heading_character_2,
            heading_character_2=heading_character_3,
            exclude_module_folders=exclude_module_folders,
        )

        rst = f"""\

.. _module_{self._module.name}:

{heading}
{heading_underline}

This document contains technical documentation for the ``{self._module.name}`` module.
{url_rst}

{overview_rst}

{registers_rst}

{submodule_rst}
"""

        return rst

    def create_rst_document(
        self, output_path: Path, exclude_module_folders: Optional[list[str]] = None
    ) -> None:
        """
        Create an ``.rst`` file in ``output_path`` with the content from :meth:`.get_rst_document`.
        If the module has registers, the HTML page will also be generated in ``output_path``, so
        that e.g. sphinx can be run directly.

        Arguments:
            output_path: Document will be placed here.
            exclude_module_folders: Folder names within the module root that shall be
                excluded from documentation.
        """
        register_list = self._module.registers
        if register_list is not None:
            HtmlPageGenerator(register_list=register_list, output_folder=output_path).create()

        rst = self.get_rst_document(exclude_module_folders=exclude_module_folders)
        create_file(output_path / f"{self._module.name}.rst", contents=rst)

    def _get_vhdl_files(
        self, exclude_files: Optional[set[Path]], exclude_folders: list[Path]
    ) -> list[Path]:
        """
        Get VHDL files that shall be included in the documentation, in order.
        """
        # Exclude all file types except VHDL.
        hdl_files = self._module.get_documentation_files(
            files_avoid=exclude_files, include_verilog=False, include_systemverilog=False
        )

        def file_should_be_included(hdl_file: "HdlFile") -> bool:
            if file_is_in_directory(hdl_file.path, exclude_folders):
                return False

            return True

        vhdl_file_paths = [
            hdl_file.path for hdl_file in hdl_files if file_should_be_included(hdl_file)
        ]

        # Sort by file name
        def sort_key(path: Path) -> str:
            return path.name

        vhdl_files = sorted(vhdl_file_paths, key=sort_key)

        return vhdl_files

    def _get_vhdl_file_rst(
        self,
        vhdl_file_path: Path,
        heading_character: str,
        heading_character_2: str,
        netlist_builds: list["VivadoNetlistProject"],
    ) -> str:
        """
        Get reStructuredText documentation for a VHDL file.
        """
        vhdl_file_documentation = VhdlFileDocumentation(vhdl_file_path)

        file_rst = vhdl_file_documentation.get_header_rst()
        file_rst = "" if file_rst is None else file_rst

        if self._repository_url:
            url_rst = (
                f"`View source code on {self._repository_name} "
                f"<{self._repository_url}/{vhdl_file_path.relative_to(self._module.path)}>`__."
            )
        else:
            url_rst = ""

        symbolator_rst = self._get_symbolator_rst(vhdl_file_documentation)
        symbolator_rst = "" if symbolator_rst is None else symbolator_rst

        entity_name = vhdl_file_path.stem

        resource_utilization_rst = self._get_resource_utilization_rst(
            entity_name=entity_name,
            heading_character=heading_character_2,
            netlist_builds=netlist_builds,
        )

        heading = f"{vhdl_file_path.name}"
        heading_underline = heading_character * len(heading)

        rst = f"""
.. _{self._module.name}.{entity_name}:

{heading}
{heading_underline}

{url_rst}

{symbolator_rst}

{file_rst}

{resource_utilization_rst}
"""

        return rst

    @staticmethod
    def _get_symbolator_rst(vhdl_file_documentation: VhdlFileDocumentation) -> str:
        """
        Get RST for rendering a symbolator component.
        """
        component = vhdl_file_documentation.get_symbolator_component()
        if component is None:
            return ""

        indent = "  "
        rst = ".. symbolator::\n\n"
        rst += indent + component.replace("\n", f"\n{indent}")

        return rst

    def _get_resource_utilization_rst(  # pylint: disable=too-many-locals,too-many-branches
        self, entity_name: str, heading_character: str, netlist_builds: list["VivadoNetlistProject"]
    ) -> str:
        # First, loop over all netlist builds for this module and assemble information
        generics = []
        checkers = []
        for netlist_build in netlist_builds:
            if netlist_build.build_result_checkers:
                generics.append(netlist_build.static_generics)

                # Create a dictionary for each build, that maps "Checker name": "value"
                checker_dict = {}
                for checker in netlist_build.build_result_checkers:
                    # Casting the limit to string yields e.g. "< 4", "4" or "> 4"
                    checker_dict[checker.name] = str(checker.limit)

                checkers.append(checker_dict)

        # Make RST of the information
        rst = ""
        if generics:
            heading = "Resource utilization"
            heading_underline = heading_character * len(heading)

            module_py_name = f"module_{self._module.name}.py"
            if self._repository_url:
                module_py_rst = f"`{module_py_name} <{self._repository_url}/{module_py_name}>`__"
            else:
                module_py_rst = f"``{module_py_name}``"

            rst = f"""
.. _{self._module.name}.{entity_name}.resource_utilization:

{heading}
{heading_underline}

This entity has `netlist builds <{WEBSITE_URL}/netlist_build.html>`__ set up with
`automatic size checkers <{WEBSITE_URL}/netlist_build.html#build-result-checkers>`__
in {module_py_rst}.
The following table lists the resource utilization for the entity, depending on
generic configuration.

.. list-table:: Resource utilization for **{entity_name}** netlist builds.
  :header-rows: 1

"""

            # Make a list of the unique checker names. Use list rather than set to preserve order.
            checker_names = []
            for build_checkers in checkers:
                for checker_name in build_checkers:
                    if checker_name not in checker_names:
                        checker_names.append(checker_name)

            # Fill in the header row
            rst += "  * - Generics\n"
            for checker_name in checker_names:
                rst += f"    - {checker_name}\n"

            # Make one row for each netlist build
            for build_idx, generic_dict in enumerate(generics):
                generic_strings = [f"{name} = {value}" for name, value in generic_dict.items()]
                generics_rst = "\n\n      ".join(generic_strings)

                rst += f"""\
  * - {generics_rst}"""

                # If the "top" of the project is different than the entity, we assume that it
                # is a wrapper. Add a note to the table about this. This occurs e.g. in the reg_file
                # and fifo modules.
                if netlist_builds[build_idx].top != entity_name:
                    if generic_strings:
                        # If there is already something in the generic column, this note shall be
                        # placed on a new line.
                        leader = "\n\n      "
                    else:
                        # Otherwise, i.e. if the netlist build has no generics set,
                        # the note shall be placed as the first thing. This is the case with
                        # two builds in the reg_file module.
                        leader = ""

                    rst += f"""\
{leader}(Using wrapper

      {netlist_builds[build_idx].top}.vhd)"""

                rst += "\n"

                for checker_name in checker_names:
                    checker_value = checkers[build_idx][checker_name]
                    rst += f"    - {checker_value}\n"

        return rst



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from pathlib import Path

# Third party libraries
import pytest

# First party libraries
from tsfpga.ip_core_file import IpCoreFile


def test_can_cast_to_string_without_error():
    str(IpCoreFile(Path("/apa/my_core.tcl")))
    str(IpCoreFile(Path("/apa/my_core.tcl"), apa=123, hest="true", zebra=False))


def test_name():
    assert IpCoreFile(Path("/apa/my_core.tcl")).name == "my_core"


def test_name_with_spaces_should_raise_exception():
    path = Path("/apa/my core.tcl")
    with pytest.raises(ValueError) as exception_info:
        # pylint: disable=expression-not-assigned
        IpCoreFile(path).name
    assert str(exception_info.value) == f"File name may not contain spaces: {path}"



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
import json
from typing import Optional, Union

# Local folder libraries
from .logic_level_distribution_parser import LogicLevelDistributionParser


class BuildResult:
    """
    Attributes:
        name (`str`): The name of the build.
        success (`bool`): True if the build and all pre- and post hooks succeeded.
        synthesis_size (`dict`): A dictionary with the utilization of primitives for the
            synthesized design.
            Will be ``None`` if synthesis failed or did not run.
        implementation_size (`dict`): A dictionary with the utilization of primitives for
            the implemented design.
            Will be ``None`` if implementation failed or did not run.
        logic_level_distribution (str): A table with logic level distribution as reported by Vivado.
            Will be ``None`` for non-netlist builds.
            Will be ``None`` if synthesis failed or did not run.
    """

    def __init__(self, name: str):
        """
        Arguments:
            name: The name of the build.
        """
        self.name = name
        self.success: bool = True
        self.synthesis_size: Optional[dict[str, int]] = None
        self.implementation_size: Optional[dict[str, int]] = None
        self.logic_level_distribution: Optional[str] = None

    def size_summary(self) -> Optional[str]:
        """
        Return a string with a formatted message of the size.

        Return:
            A human-readable message of the latest size. ``None`` if no size is set.
        """
        build_step = None
        size = None

        if self.implementation_size:
            build_step = "implementation"
            size = self.implementation_size

        elif self.synthesis_size:
            build_step = "synthesis"
            size = self.synthesis_size

        else:
            return None

        return f"Size of {self.name} after {build_step}:\n{json.dumps(size, indent=2)}"

    def report(self) -> Optional[str]:
        """
        Return a report of the build result. Includes all metrics and information that has been
        extracted from the Vivado reports.
        """
        result = self.size_summary()

        if self.logic_level_distribution:
            result = f"{result}\nLogic level distribution:\n{self.logic_level_distribution}"

        return result

    @property
    def maximum_logic_level(self) -> Union[None, int]:
        """
        The maximum level in the the :attr:`.BuildResult.logic_level_distribution`.
        Will be ``None`` for non-netlist builds.
        Will be ``None`` if synthesis failed or did not run.

        Return:
            The maximum logic level.
        """
        if not self.logic_level_distribution:
            return None

        return LogicLevelDistributionParser.get_maximum_logic_level(
            table=self.logic_level_distribution
        )



# --------------------------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
#
# This file is part of the tsfpga project, a project platform for modern FPGA development.
# https://tsfpga.com
# https://github.com/tsfpga/tsfpga
# --------------------------------------------------------------------------------------------------

# Standard libraries
from enum import Enum, auto
from pathlib import Path


class HdlFile:
    """
    Class for representing a HDL source code file in the file system.
    """

    class Type(Enum):
        """
        Enumeration of supported HDL file types.
        """

        VHDL = auto()
        VERILOG_SOURCE = auto()
        VERILOG_HEADER = auto()
        SYSTEMVERILOG_SOURCE = auto()
        SYSTEMVERILOG_HEADER = auto()

    # Decides which file endings are associated with which file type.
    file_endings_mapping = {
        Type.VHDL: (".vhd", ".vhdl"),
        Type.VERILOG_SOURCE: (".v",),
        Type.VERILOG_HEADER: (".vh",),
        Type.SYSTEMVERILOG_SOURCE: (".sv",),
        Type.SYSTEMVERILOG_HEADER: (".svh",),
    }

    # A tuple of all supported HDL file endings.
    file_endings = tuple(
        file_ending
        for type_file_endings in file_endings_mapping.values()
        for file_ending in type_file_endings
    )

    def __init__(self, path: Path) -> None:
        """
        Arguments:
            path: Path to a HDL source code  file.
        """
        self._path = path

        for file_type, file_endings in self.file_endings_mapping.items():
            if path.name.endswith(file_endings):
                self._type = file_type
                break
        else:
            raise ValueError(f"Unsupported HDL file ending: {path}")

    @property
    def path(self) -> Path:
        """
        Path to the HDL file.
        Getter for read-only class variable.
        """
        return self._path

    @property
    def type(self) -> Type:
        """
        The file type of the HDL file.
        Getter for read-only class variable.
        """
        return self._type

    def __str__(self) -> str:
        return f"{self.__class__.__name__}('{self._path}', '{self._type}')"

    def __repr__(self) -> str:
        return str(self)
